"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const globallyBeforeEachStep_1 = require("./globallyBeforeEachStep");
const getGlobalTestMethods_1 = require("./lib/getGlobalTestMethods");
const parseFeature_1 = require("./lib/parseFeature");
/**
 * Create a gherkin test based on a feature file
 */
function Feature(config, configure) {
    const { feature, methods = getGlobalTestMethods_1.getGlobalTestMethods(), defaultTimeout } = (typeof config === 'string'
        ? { feature: config }
        : config);
    const { featureBuilder, ast } = parseFeature_1.parseFeature({ feature, stackIndex: 2 });
    describeFeature({ ast, configure, featureBuilder, methods, defaultTimeout });
    return { ast, featureBuilder };
}
exports.Feature = Feature;
exports.GherkinTest = Feature;
/** Ensures correct synchronous execution of the configuration stage */
function configureMethods({ configure, featureBuilder }) {
    let resolve;
    const onConfigured = (callback) => {
        resolve = callback;
    };
    const methods = {
        Scenario: featureBuilder.Scenario(),
        ScenarioOutline: featureBuilder.ScenarioOutline(onConfigured),
        Background: featureBuilder.Background(),
        BeforeEach: featureBuilder.BeforeEach(),
        BeforeAll: featureBuilder.BeforeAll(),
        AfterEach: featureBuilder.AfterEach(),
        AfterAll: featureBuilder.AfterAll(),
    };
    configure(methods);
    if (resolve) {
        resolve();
    }
}
/** Intruments the gherkin feature test in the test framework */
function describeFeature({ featureBuilder, ast, configure, methods, defaultTimeout }) {
    let error;
    const title = formatTitle(Object.assign({}, ast.feature, { prefix: 'Feature:' }));
    methods.describe(title, () => {
        try {
            configureMethods({ configure, featureBuilder });
            const { background, scenarios, beforeAll: beforeAllHooks, afterAll: afterAllHooks, afterEach: afterEachHooks, beforeEach: beforeEachHooks, } = featureBuilder.feature;
            afterAllHooks.forEach(({ fn, timeout = defaultTimeout }) => { methods.afterAll(fn, timeout); });
            beforeAllHooks.forEach(({ fn, timeout = defaultTimeout }) => { methods.beforeAll(fn, timeout); });
            scenarios.forEach((scenario) => {
                describeScenario({
                    background, scenario,
                    afterEachHooks, beforeEachHooks,
                    methods, defaultTimeout,
                    initialState: undefined,
                });
            });
        }
        catch (e) {
            error = e;
        }
    });
    if (error) {
        throw error;
    }
}
/** Executes a step function and ensures state passing */
function executeStep(step, initialState) {
    return __awaiter(this, void 0, void 0, function* () {
        const state = yield step.fn(initialState, ...step.params);
        if (state !== undefined) {
            return state;
        }
        return initialState;
    });
}
/** Instruments the tests for a scenario in the test framework */
function describeScenario({ initialState, background, scenario, beforeEachHooks, afterEachHooks, methods, defaultTimeout, }) {
    const title = formatTitle(Object.assign({}, scenario.gherkin, { prefix: `Scenario:` }));
    const { skip, only } = scenario;
    const describe = narrowTestMethod(methods.describe, { skip, only });
    describe(title, () => {
        if (!skip) {
            afterEachHooks.forEach(({ fn, timeout = defaultTimeout }) => { methods.afterAll(fn, timeout); });
            beforeEachHooks.forEach(({ fn, timeout = defaultTimeout }) => { methods.beforeAll(fn, timeout); });
        }
        const steps = {
            Background: background && background.Given,
            Given: scenario.Given,
            When: scenario.When,
            Then: scenario.Then,
        };
        describeGherkinOperations({ steps, state: initialState, methods });
    });
}
/** Instruments step functions in the test framework */
function describeGherkinOperations({ state, methods, defaultTimeout, steps: { Given, Then, When, Background = new Map(), }, }) {
    /** Used to determine if previous dependent steps have failed */
    let aStepHasFailed = false;
    let reducedState = state;
    /** Curried fn which assigns a step function to the test framework method 'test' */
    function AssignStep(prefix = '') {
        return (step) => {
            const title = formatTitle({ name: step.name, prefix });
            const timeout = step.timeout || defaultTimeout;
            const testFn = () => __awaiter(this, void 0, void 0, function* () {
                /** If a previous step failed, just return so we dont get bad execution errors */
                if (aStepHasFailed) {
                    return;
                }
                try {
                    /** Execute the globalBeforeEachStepHooks before the step */
                    if (globallyBeforeEachStep_1.globalBeforeEachStepHooks.length) {
                        reducedState = globallyBeforeEachStep_1.globalBeforeEachStepHooks.reduce((nextState, callback) => {
                            return callback(step, nextState, ...step.params);
                        }, reducedState);
                    }
                    reducedState = yield executeStep(step, reducedState);
                    return reducedState;
                }
                catch (error) {
                    aStepHasFailed = true;
                    throw error;
                }
            });
            const testMethod = step.skip ? methods.test.skip : methods.test;
            testMethod(title, testFn, timeout);
        };
    }
    Background.forEach(AssignStep('(Background) Given:'));
    Given.forEach(AssignStep('Given:'));
    When.forEach(AssignStep('When:'));
    Then.forEach(AssignStep('Then:'));
}
/** Resolves to the correct skipping test method based on configuration options */
function narrowTestMethod(method, { skip, only }) {
    if (only) {
        return method.only;
    }
    if (skip) {
        return method.skip;
    }
    return method;
}
/** Formats a test title based on tags, name and description */
function formatTitle({ name, tags: inputTags, description = '', prefix = '', }) {
    const leftPadLines = (str, pad = '  ') => str.split('\n').map((line) => `${pad}${line}`).join('\n');
    const tags = inputTags
        ? ` ${inputTags.map((tag) => tag.name).join(' ')}`
        : '';
    const title = `${prefix && `${prefix} `}${name}${tags}`;
    return [
        title,
        `${leftPadLines(description)}`,
    ]
        .filter(Boolean)
        .join('\n').trimRight();
}
