"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Interpolator = require("trans-interpolator");
const GherkinTableReader_1 = require("./lib/GherkinTableReader");
const matchInGherkinCollection_1 = require("./lib/matchInGherkinCollection");
const parseGherkinParameters_1 = require("./lib/parseGherkinParameters");
/**
 * This class populates the `feature` property with
 * a data structure that is mapped to a gherkin feature file.
 *
 * The feature file is built up as Scenario()() etc. are invoked,
 * causing the gherkin AST to be parsed and mapped up to the invoking function.
 */
class FeatureBuilder {
    constructor(ast) {
        this.feature = {
            scenarios: [],
            scenarioOutlines: [],
            afterAll: [],
            beforeAll: [],
            afterEach: [],
            beforeEach: [],
        };
        this.feature.gherkin = ast.feature;
    }
    Scenario() {
        const { gherkin: { children: collection } } = this.feature;
        const build = (match) => {
            const gherkin = matchInGherkinCollection_1.matchInGherkinCollection({
                collection, match,
                matchProperty: 'name',
                type: 'Scenario',
            });
            return ScenarioFluidBuilder({ match, gherkin, feature: this.feature.gherkin });
        };
        const EntryPoint = (overrides = {}) => {
            return (match) => {
                const { definition, steps } = build(match);
                this.feature.scenarios = [...this.feature.scenarios, Object.assign({}, definition, overrides)];
                return steps;
            };
        };
        const fn = EntryPoint();
        fn.skip = EntryPoint({ skip: true });
        fn.only = EntryPoint({ only: true });
        return fn;
    }
    ScenarioOutline(onConfigured) {
        const { gherkin: { children: collection } } = this.feature;
        const build = (match) => {
            const gherkin = matchInGherkinCollection_1.matchInGherkinCollection({
                collection, match,
                matchProperty: 'name',
                type: 'ScenarioOutline',
            });
            return ScenarioOutlineFluidBuilder({ match, gherkin, onConfigured, feature: this.feature.gherkin });
        };
        const EntryPoint = (overrides = {}) => {
            return (match) => {
                const { scenarioOutline, scenarios, steps } = build(match);
                this.feature.scenarios = [
                    ...this.feature.scenarios,
                    ...scenarios.map((scenario) => (Object.assign({}, scenario, overrides))),
                ];
                this.feature.scenarioOutlines = [
                    ...this.feature.scenarioOutlines,
                    scenarioOutline,
                ];
                return steps;
            };
        };
        const fn = EntryPoint();
        fn.skip = EntryPoint({ skip: true });
        fn.only = EntryPoint({ only: true });
        return fn;
    }
    Background() {
        const { gherkin: { children: collection } } = this.feature;
        const build = (match = '') => {
            const gherkin = matchInGherkinCollection_1.matchInGherkinCollection({
                collection, match,
                matchProperty: 'name',
                type: 'Background',
            });
            return BackgroundFluidBuilder({ match, gherkin, feature: this.feature.gherkin });
        };
        const fn = (match = '') => {
            const { definition, steps } = build(match);
            this.feature.background = definition;
            return steps;
        };
        return fn;
    }
    AfterAll() {
        return (fn, options = {}) => {
            this.feature.afterAll = [...this.feature.afterAll, Object.assign({ fn }, options)];
        };
    }
    BeforeAll() {
        return (fn, options = {}) => {
            this.feature.beforeAll = [...this.feature.beforeAll, Object.assign({ fn }, options)];
        };
    }
    AfterEach() {
        return (fn, options = {}) => {
            this.feature.afterEach = [...this.feature.afterEach, Object.assign({ fn }, options)];
        };
    }
    BeforeEach() {
        return (fn, options = {}) => {
            this.feature.beforeEach = [...this.feature.beforeEach, Object.assign({ fn }, options)];
        };
    }
}
exports.FeatureBuilder = FeatureBuilder;
/**
 * A function is which returns a chainable interface
 * and instruments step definitions into a store when executed
 */
function FluidFn({ fluid, collectionParams, store, definition }) {
    const fluidFn = ((match, fn = identity, options = {}) => {
        const gherkin = matchInGherkinCollection_1.matchInGherkinCollection(Object.assign({ match }, collectionParams));
        store.set(match, Object.assign({ fn, gherkin, definition, name: gherkin[collectionParams.matchProperty], params: parseGherkinParameters_1.parseGherkinParameters(gherkin, match), skip: fn === identity }, options));
        return fluid;
    });
    fluidFn.skip = (match, fn, options = {}) => fluidFn(match, fn, Object.assign({}, options, { skip: true }));
    return fluidFn;
}
/**
 * Lazy Fluid Function. Only stores the `match` and `fn`
 */
function LazyFluidFn({ fluid, store }) {
    const fluidFn = ((match, fn = identity, options = {}) => {
        store.set(match, Object.assign({ fn, skip: fn === identity }, options));
        return fluid;
    });
    fluidFn.skip = (match, fn, options = {}) => fluidFn(match, fn, Object.assign({}, options, { skip: true }));
    return fluidFn;
}
function ScenarioFluidBuilder({ match, gherkin, feature, FluidFnFactory = FluidFn }) {
    const { steps: collection } = gherkin;
    const definition = {
        match, gherkin, feature,
        name: gherkin.name,
        Given: new Map(),
        Then: new Map(),
        When: new Map(),
    };
    const thenFluid = {};
    const Then = FluidFnFactory({
        definition,
        fluid: thenFluid,
        store: definition.Then,
        collectionParams: { collection, matchProperty: 'text', type: 'Step' },
    });
    const whenFluid = { Then };
    const When = FluidFnFactory({
        definition,
        fluid: whenFluid,
        store: definition.When,
        collectionParams: { collection, matchProperty: 'text', type: 'Step' },
    });
    const givenFluid = { Then, When };
    const Given = FluidFnFactory({
        definition,
        fluid: givenFluid,
        store: definition.Given,
        collectionParams: { collection, matchProperty: 'text', type: 'Step' },
    });
    // Apply conjuctions via mutation
    thenFluid.And = Then;
    whenFluid.And = When;
    givenFluid.And = Given;
    return {
        definition,
        steps: { Given, Then, When },
    };
}
function ScenarioOutlineFluidBuilder({ match, gherkin, onConfigured, feature }) {
    const { examples: [exampleSrc], steps: outlineSteps, } = gherkin;
    const examples = GherkinTableReader_1.GherkinTableReader({
        rows: [exampleSrc.tableHeader, ...exampleSrc.tableBody || []],
    }).rows.mapByTop();
    const scenarioBuilders = examples.map((example, index) => {
        const interp = new Interpolator(example, { open: '<', close: '>' });
        const scenarioSteps = outlineSteps.map((step) => {
            return Object.assign({}, step, { text: interp.interpolate(step.text) });
        });
        return ScenarioFluidBuilder({
            match, feature,
            gherkin: {
                type: 'Scenario',
                keyword: 'Scenario',
                description: gherkin.description,
                steps: scenarioSteps,
                location: gherkin.location,
                name: `${gherkin.name} (${index})`,
                tags: gherkin.tags,
            },
        });
    });
    const scenarios = scenarioBuilders.map(({ definition }) => definition);
    /** This is used to make the step methods DRY */
    const scenarioBuilderSkeleton = ScenarioFluidBuilder({ match, gherkin, feature, FluidFnFactory: LazyFluidFn });
    const { steps } = scenarioBuilderSkeleton;
    const scenarioOutline = Object.assign({}, scenarioBuilderSkeleton.definition, { gherkin, match, name: gherkin.name, skip: false, only: false });
    const methods = ['Given', 'When', 'Then'];
    /**
     * When configured:
     *   - Invoke our Scenarios with the ScenarioOutline step definitions
     */
    onConfigured(() => {
        methods.forEach((methodName) => {
            const operations = scenarioOutline[methodName];
            if (!operations) {
                return;
            }
            operations.forEach(({ fn }, operationMatch) => {
                scenarioBuilders.forEach(({ steps: stepMethods }) => {
                    const method = stepMethods[methodName];
                    method(operationMatch, fn);
                });
            });
        });
    });
    return { scenarioOutline, scenarios, steps };
}
function BackgroundFluidBuilder({ match, gherkin, feature }) {
    const { steps: collection } = gherkin;
    const definition = {
        gherkin, match, feature,
        name: gherkin.name,
        Given: new Map(),
        skip: false,
        only: false,
    };
    const givenFluid = {};
    const Given = FluidFn({
        definition,
        fluid: givenFluid,
        store: definition.Given,
        collectionParams: { collection, matchProperty: 'text', type: 'Step' },
    });
    givenFluid.And = Given;
    return {
        definition,
        steps: { Given },
    };
}
const identity = (v) => v;
